/* $Header: /home/vikas/src/xtacacsd/RCS/taclast.c,v 1.11 1997/01/27 10:44:29 vikas Exp $ */

/*
 * This is similar to the 'last' program on Unix systems but for processing
 * my xtacacsd utmp/wtmp files (xtacacsd runs on Cisco routers).
 * Also acts like a simple 'who' program to read the data in the tacacs UTMP
 * file.
 *
 * Note that some login/logout records might be missing. In such a case:
 *	LOGIN - LOGOUT  = normal
 *	LOGIN - LOGIN   = LOGIN - (LOGOUT)LOGIN
 *	LOGOUT - LOGIN  = still logged on
 *	LOGOUT - LOGOUT = LOGOUT(LOGIN) - LOGOUT (missing first char)
 *
 * BUGS:
 *	While reading an ASCII wtmp file, it converts the read data into
 * the fixed length tacutmp structure. If the read data is larger then the
 * conversion might truncate the data.
 * Can avoid by setting the xx_NAMESIZE to be large from the beginning in
 * the tacutmp.h  header file.
 *
 * AUTHOR:
 *  	Vikas Aggarwal, vikas@navya.com, June 1995
 *
 * Partially derived from:
 *	$ NetBSD: last.c,v 1.6 1994/12/24 16:49:02 cgd Exp
 *
 * $Log: taclast.c,v $
 * Revision 1.11  1997/01/27 10:44:29  vikas
 * Small error in printing out the "remaining" entries (reached EoFile
 * but did not see login entry).
 *
 * Revision 1.10  1997/01/11 10:54:45  vikas
 * Now sees if hostname on logout is NULL (as on freeBSD)
 * Also prints out all users that dont have a login when the
 * end of the wtmp file is reached.
 * Tested version.
 *
 * Revision 1.9  1997/01/09 11:05:26  vikas
 * Now using the islogout() function. Also handling the case if
 * we get 2 consecutive logout's.
 *
 * Revision 1.8  1997/01/08 10:52:26  vikas
 * Did not need the & in the structure elements during sscandf()
 *
 * Revision 1.7  1997/01/07 10:36:45  vikas
 * (older logs are in the RCS directory)
 * -Added code to handle ASCII wtmp files fetch_tacutmp_ascii()
 * -Separated printf's in the wtmp code
 * -Now keeps and prints the incorrect times in the totals
 *
 *
 */


#ifndef lint
static char rcsid[] = "$RCSfile: taclast.c,v $ $Revision: 1.11 $ $Date: 1997/01/27 10:44:29 $" ;
#endif /* not lint */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>		/* for MAXHOSTNAMELEN on SCO ? */
#include <sys/stat.h>

#include <fcntl.h>
#include <sys/file.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
/* #include <tzfile.h>			/* for SECSPERDAY, etc. */
#include <unistd.h>			/* for SEEK_ definitions */

#include "tacutmp.h"			/* our very own utmp.h file */

#ifndef	SECSPERDAY
#define	SECSPERDAY	((long) (60 * 60 * 24))
#endif
#ifndef DAYSPERWEEK
#define DAYSPERWEEK	7
#endif

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 128
#endif

#if defined(SYSV) || defined(SVR4) || defined(__svr4__)
#  define index strchr
#  define bcopy(a, b, c) memmove(b, a, c)
#  define bzero(a, b)    memset(a, 0, b)
#endif

#define	NO	0				/* false/no */
#define	YES	1				/* true/yes */

/* Following structure is used to gather time information */
typedef struct arg {
  char	  *name;				/* argument */
  time_t  errsecs;				/* inaccurate seconds */
  time_t  oksecs;				/* confirmed seconds  */
#define	HOST_TYPE	-2
#define	TTY_TYPE	-3
#define	USER_TYPE	-4
	struct arg	*next;			/* linked list pointer */
} ARG;
ARG	*HOSTarglist, *USERarglist;		/* head of linked list */

typedef struct ttytab {
  time_t	logout;				/* log out time */
  int	waslogin;				/* boolean for prev entry */
  char	tty[UT_LINESIZE + 1];			/* terminal name */
  char	host[UT_HOSTSIZE + 1];			/* host name */
  char	name[UT_NAMESIZE + 1];			/* user on this TTY */
  struct ttytab	*next;				/* linked list pointer */
} TTY;
TTY	*ttylist;				/* head of linked list */

#ifndef WTMP
#define WTMP	"/var/adm/wtmp"
#define UTMP	"/etc/utmp"
#endif

static int	asciiwtmp;			/* use ascii WTMP file */
static int	whomode;			/* do a 'who', not 'last' */
static int	ignorehost= 0;			/* dont compare hostnames */
static int	dototals;			/* want totaltimes */
static int	totalsonly;			/* want totaltimes ONLY */
static int	noargs;				/* no cmd line args */
static long	maxrec;				/* records to display */
static char	*file;				/* wtmp file */
static char	*weekdays[] = {"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", 
			       "Sat", "Sun"};

/* Following are not really needed, but are external variables required
 * in the uwtmp.c module.
 */
int	wtmpfd, utmpfd;
int 	debug, logging, update_host_wtmp, stand, interruptible;
char	*from_hname, *wtmp_file, *utmp_file;

void	 addarg (/* int, char * */);
TTY	*addtty (/* char * */);
void	 hostconv (/* char * */);
char	*ttyconv (/* char * */);
int	 want (/* struct tacutmp *, int */);
void	 wtmp (/* void */);
time_t	 pr_entry(/* struct tacutmp *, TTY * */);

char	*prognm;

main(argc, argv)
  int argc;
  char *argv[];
{
  extern int optind;
  extern char *optarg;
  int ch;
  char *p;

  prognm = argv[0];	/* save program name */
  if (!strcmp(prognm, "tacwho"))		/* if called as tacwho */
    whomode = 1;

  maxrec = -1;
  while ((ch = getopt(argc, argv, "0123456789af:h:itw")) != EOF)
    switch (ch) {
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      /*
       * kludge: last was originally designed to take
       * a number after a dash.
       */
      if (maxrec == -1) {
	p = argv[optind - 1];
	if (p[0] == '-' && p[1] == ch && !p[2])
	  maxrec = atol(++p);
	else
	  maxrec = atol(argv[optind] + 1);
	if (!maxrec)
	  exit(0);
      }
      break;
    case 'a':
      asciiwtmp++;
      break;
    case 'f':
      file = optarg;
      break;
    case 'h':
      /*			hostconv(optarg);	/* not used */
      addarg(HOST_TYPE, optarg);
      break;
    case 'i':
      ignorehost++;		/* if separate wtmp files */
      break;
    case 't':
      dototals++;		/* if 2, then only print totals */
      if (dototals >= 2) totalsonly++;
      break;
    case 'w':
      whomode++;
      break;
    case '?':
    default:
      (void)fprintf(stderr,
		    "usage: %s [-# (max num records)] [-f file] [-a ascii-wtmp-file] [-h hostname] [-i (for ignore host)]", prognm);
      (void)fprintf(stderr,
		    " [-w (for WHO mode)] [-t (totaltimes) [-t (only totals)]] [user ...]\n");
      exit(1);
    }

  if (argc) {
#ifndef POSIX
    setlinebuf(stdout);	/* */
#else
    setbuf(stdout, NULL);	/* */
#endif
    for (argv += optind; *argv; ++argv) {
      addarg(USER_TYPE, *argv);
    }
  }

  if (!HOSTarglist && !USERarglist)
    noargs = 1;			/* no cmd line arguments */

  if (whomode) doutmp() ;
  else  wtmp();

  exit(0);
}	/* end main() */


/*+
 * Return one filled in tacutmp structure. Return NULL on end of file.
 * Work backwards in the file.
 * Works on ASCII wtmp files.
 */
struct tacutmp *fetch_tacutmp_ascii()
{
  static struct tacutmp entry;
  static char buf[4096];
  static int firsttime = 1, bytes, wfd;
  static long bl;
  
  if (firsttime)
  {
    struct stat	stb;			/* stat of file for size */

    firsttime = 0;
    if (file == (char *)NULL || *file == '\0')
      file = WTMP ;

    if ((wfd = open(file, O_RDONLY, 0)) < 0 || fstat(wfd, &stb) == -1) {
      fprintf(stderr, "%s: %s: ", prognm, file);
      perror ((char *)NULL);
      exit (1);
    }
    /* extract total # of complete reads */
    bl = (stb.st_size + sizeof(buf) - 1) / sizeof(buf);
  }

  /* work backwards in the file */
fetchmore:
  if (bl > 0 || bytes > 0)
  {
    register char *p;
    static int getmore = 0;

    if (bytes == 0 || getmore)
    {
      --bl;
      getmore = 0;
      if (lseek(wfd, (off_t)((bl * sizeof(buf)) + bytes), SEEK_SET) == -1 ||
	  (bytes = read(wfd, buf, sizeof(buf))) == -1)
      {
	fprintf(stderr, "%s: %s: ", prognm, file);
	perror((char *)NULL);
	exit (1);
      }

      /* make sure block ends on a \n */
      for (p = buf + bytes ; p > buf && *p != '\n' ; --p)
	;
      if (*p == '\n') 	bytes = p - buf;
      else  ++bytes;	/* to insert a NULL in this line */
    }

    *(buf + bytes) = '\0';
    for (p = buf + bytes ; p > buf && *p != '\n'; --p)
      ;

    if (*p == '\n' || bl == 0)
    {
      bytes = p - buf;		/* decrement */
      if (*p == '\n')
	++p;
#ifdef XTACUTMP
      sscanf(p, "%ld %s %s %s %s", entry.ut_time, entry.ut_host,
	     entry.ut_line, entry.ut_name, entry.ut_comment);
#else
      sscanf(p, "%ld %s %s %s", entry.ut_time, entry.ut_host,
	     entry.ut_line, entry.ut_name);
#endif
      return (&entry);	/*  */
      /*      fprintf(stderr, "%s\n", p);	/* for debugging */
    }
    else		/* need to scan more data */
    {
      getmore = 1;
      goto fetchmore;
    }
  }	/* if (bl > 0) */
  
  /* Here if end of file */
  return(NULL);
}	/* fetch_tacutmp_ascii() */


/*+
 * Return one filled in tacutmp structure. Return NULL on end of file.
 * Work backwards in the file.
 * This is for binary fixed record length wtmp files.
 */
struct tacutmp *fetch_tacutmp()
{
  static struct tacutmp  tacbuf[1024];		/* utmp read buffer */
  static int  bytes, wfd;
  static int  firsttime = 1;
  static int cur = 0;			/* current tacbuf number */
  static long bl;

  if (asciiwtmp)
    return(fetch_tacutmp_ascii());

  if (firsttime)
  {
    struct stat	stb;			/* stat of file for size */

    firsttime = 0;
    if (file == (char *)NULL || *file == '\0')
      file = WTMP ;

    if ((wfd = open(file, O_RDONLY, 0)) < 0 || fstat(wfd, &stb) == -1) {
      fprintf(stderr, "%s: %s: ", prognm, file);
      perror ((char *)NULL);
      exit (1);
    }

    if ( (stb.st_size % sizeof tacbuf[0]) != 0) {
      fprintf(stderr,
	      "ERROR: WTMP record sizes not compatible with compiled tacutmp structure.\n");
      fprintf(stderr,
	      "This is controlled using the XTACUTMP compile time define.\n");
#ifdef XTACUTMP
      fprintf(stderr, "Recompile program without XTACUTMP.\n");
#else
      fprintf(stderr, "Recompile program with XTACUTMP.\n");
#endif
      exit(1);
    }

    /* extract total # of reads */
    bl = (stb.st_size + sizeof(tacbuf) - 1) / sizeof(tacbuf);

  }	/* end if (firsttime) */

  /* work backwards in the file */
  if (bl > 0 || cur > 0)
  {
    if (cur == 0)
    {			/* read in a new block of data */
      --bl;
      if (lseek(wfd, (off_t)(bl * sizeof(tacbuf)), SEEK_SET) == -1 ||
	  (bytes = read(wfd, tacbuf, sizeof(tacbuf))) == -1)
      {
	fprintf(stderr, "%s: %s: ", prognm, file);
	perror((char *)NULL);
	exit (1);
      }
      cur = bytes / sizeof(tacbuf[0]);
    }

    return (&tacbuf[--cur]);
  }

  return(NULL);
}	/* end fetch_tacutmp() */

/*
 * wtmp --
 *	read through the wtmp file
 */
void
wtmp()
{
  struct tacutmp  *bp;				/* current structure */
  TTY	*T;					/* tty list entry */
  time_t start, delta;				/* time difference */
  time_t prevtm, stoptime ;			/* prev entry time */
  char	*ct, *crmsg;

  time(&start);			/* store start time */
  stoptime = 0;

  while ( (bp = fetch_tacutmp()) != NULL)
  {
    prevtm = bp->ut_time;	/* need this at the end... */
    /*
     * if the terminal line is '~', the machine stopped.
     */
    if (bp->ut_line[0] == '~' && !bp->ut_line[1]) {
      /* everybody just logged out. Store timestamp as negative */
      for (T = ttylist; T; T = T->next) {
	if (ignorehost || strncasecmp(T->host, bp->ut_host, UT_HOSTSIZE) == 0 )
	  T->logout = -bp->ut_time;	/* indicate stopped entry */
      }
      stoptime = -bp->ut_time;	/* used in addTTY */
      crmsg = strncmp(bp->ut_name, "shutdown",
		      UT_NAMESIZE) ? "crash" : "shutdown";
      if (want(bp, NO)) {
	ct = ctime(&bp->ut_time);
	printf("%-*.*s  %-*.*s %-*.*s %10.10s %5.5s \n",
	       UT_NAMESIZE, UT_NAMESIZE, bp->ut_name,
	       UT_LINESIZE, UT_LINESIZE, bp->ut_line,
	       UT_HOSTSIZE, UT_HOSTSIZE, bp->ut_host,
	       ct, ct + 11);
	if (maxrec != -1 && !--maxrec)
	  return;
      }
      continue;
    }	/* end if machine stopped */

    /*
       * if the line is '{' or '|', date got set; see
       * utmp(5) for more info.
       */
    if ((bp->ut_line[0] == '{' || bp->ut_line[0] == '|')
	&& !bp->ut_line[1]) {
      if (want(bp, NO)) {
	ct = ctime(&bp->ut_time);
	printf("%-*.*s  %-*.*s %-*.*s %10.10s %5.5s \n",
	       UT_NAMESIZE, UT_NAMESIZE, bp->ut_name,
	       UT_LINESIZE, UT_LINESIZE, bp->ut_line,
	       UT_HOSTSIZE, UT_HOSTSIZE, bp->ut_host,
	       ct, ct + 11);
	if (maxrec && !--maxrec)
	  return;
      }
      continue;
    }	/* end if date got set */

    /* find associated tty in TTY list */
    for (T = ttylist;; T = T->next) {
      if (!T) {
	/* add new one */
	T = addtty(stoptime, bp->ut_line, bp->ut_host, bp->ut_name);
	break;
      }
      if (!strncmp(T->tty, bp->ut_line, UT_LINESIZE))	/* line matches */
      {
	if (ignorehost)
	  break;
	/* Some hosts don't log a hostname on logouts */
	if (*(T->host) == '\0' ||
	    strncasecmp(T->host, bp->ut_host, UT_HOSTSIZE) == 0)
	  break;
      }
    }

    /* if this is a login entry */
    if (!islogout(bp) && want(bp, YES))
    {
      int isaccurate = 0;	/* assume incorrect */

      /*
       * Check for record correctness- if prev was a logout and:
       *	if prev username == NULL (old style logouts)
       *     OR	if prev username == LOGOUT_CHAR + curr name
       *     OR if prev username == curr name
       */
      if (!T->waslogin)
	if (*(T->name) == '\0' ||
	    (*(T->name) == UT_LOGOUT_CHAR &&
	     !strncmp((bp->ut_name)+1, (T->name)+1, sizeof(bp->ut_name)-1)) ||
	    !strncmp(bp->ut_name, T->name, sizeof(bp->ut_name)) )
	  isaccurate = 1;

      delta = pr_entry(bp, T, isaccurate, crmsg);

      if (dototals)  addsecs(bp, delta, isaccurate);
      if (maxrec != -1 && !--maxrec)
	return;
    }	/* end if login entry */
    else
    {	/* we got a logout entry. Check if previous also a logout... */
      if (!T->waslogin && bp->ut_name[0])
      {		/* prev also a logout */
	/*
	 * If this username is same as previous, then we CAN assume this is
	 * an extra logout and skip. However, it is more logical to
	 * flag this entire entry as inaccurate and simply
	 * assume that the prev user logged in right when this user
	 * logged out.
	 * We might be missing the first character of this user in older
	 * wtmp data files...
	 */
	delta = pr_entry(bp, T, /*isaccurate*/0, crmsg);
	if (dototals)  addsecs(bp, delta, /*isaccurate*/0); /* */
      }	/* end if (T->waslogin) */

    }	/* end if else (login) */

    /* 	We always store this bp->ut_time. If was login entry, then
     *	serves as the logout in case the logout entry is lost for the
     *  previous person
     */
    T->logout = bp->ut_time;
    /* Update this at the very end since its earlier value was required */
    if (islogout(bp))    T->waslogin = 0 ;
    else  T->waslogin = 1;

  }	/* end while(bp != NULL) */

  /* At end of the wtmp file, close out all LOGOUT's in the TTY list */
  for (T = ttylist; T != NULL; T = T->next)
    if (!T->waslogin)	/* logout entry */
    {
      struct tacutmp tbp;
      strncpy(tbp.ut_name, T->name, UT_NAMESIZE);
      strncpy(tbp.ut_line, T->tty, UT_LINESIZE);
      strncpy(tbp.ut_host, T->host, UT_HOSTSIZE);
#ifdef XTACUTMP
      strncpy(tbp.ut_comment, "Login before start of file", UT_COMMENTSIZE);
#endif
      tbp.ut_time = prevtm;
      
      delta = pr_entry(&tbp, T, /*isaccurate*/0, "");
      if (dototals)  addsecs(&tbp, delta, /*isaccurate*/0);
    }

  /* print out the total times */
  if (dototals)  pr_totaltimes();

  ct = ctime(&start);
  printf("\nwtmp processing began %10.10s %5.5s \n", ct, ct + 11);

}	/* end: wtmp() */


/*
 * want --
 *	see if want this entry (in case some hosts or username were
 * specified on the command line).
 */
int
want(bp, check)
  struct tacutmp *bp;
  int check;		/* check if username is 'ftp' or 'uucp' */
{
  ARG *step;

  if (check)
    /*
       * when uucp and ftp log in over a network, the entry in
       * the utmp file is the name plus their process id.  See
       * etc/ftpd.c and usr.bin/uucp/uucpd.c for more information.
       */
    if (!strncmp(bp->ut_line, "ftp", sizeof("ftp") - 1))
      bp->ut_line[3] = '\0';
    else if (!strncmp(bp->ut_line, "uucp", sizeof("uucp") - 1))
      bp->ut_line[4] = '\0';

  if (noargs)
    return (YES);
    
  for (step = HOSTarglist; step; step = step->next)
    if (!strncasecmp(step->name, bp->ut_host, UT_HOSTSIZE))
      return (YES);

  for (step = USERarglist; step; step = step->next)
    if (!strncmp(step->name, bp->ut_name, UT_NAMESIZE))
      return (YES);
    
  return (NO);

}

/*
 * addarg --
 *	add an entry to a linked list of arguments (used if a list of
 * username or hostname is specified on the command line or  we need
 * to add to these linked lists).
 */
void
addarg(type, arg)
  int type;
  char *arg;
{
  ARG *cur, **whichlist;

  if (!(cur = (ARG *)malloc((u_int)sizeof(ARG)))) {
    fprintf(stderr, "%s: malloc failure.\n", prognm);
    exit (1);
  }
  switch (type) {
  case HOST_TYPE:
    whichlist = &HOSTarglist;	
    break;
  case USER_TYPE:
    whichlist = &USERarglist;
    break;
  default:
    fprintf(stderr, "addarg: unknown type %d\n", type);
  }
  cur->next = *whichlist;
  cur->errsecs = 0;
  cur->oksecs = 0;
  cur->name = arg;
  *whichlist = cur;
}

/*
 * addtty --
 *	add an entry to a linked list of ttys
 */
TTY *
addtty(stoptime, ttyname, hostname, username)
  time_t stoptime;
  char *ttyname, *hostname, *username;
{
  TTY *cur;

  if (!(cur = (TTY *)malloc((u_int)sizeof(TTY)))) {
    fprintf(stderr, "%s: malloc failure.\n", prognm);
    exit (1);
  }
  cur->next = ttylist;
  cur->logout = stoptime;
  cur->waslogin = 0;	/* initialization- logout */
  bcopy(ttyname, cur->tty, UT_LINESIZE);
  bcopy(hostname, cur->host, UT_HOSTSIZE);
  bcopy(username, cur->name, UT_NAMESIZE);
  return (ttylist = cur);
}

/*
 * Add the delta seconds to the total number of seconds. Add to the arglist
 * if this is a new host or name and nothing was supplied on the command line.
 */
addsecs(bp, delta, isaccurate)
  struct tacutmp *bp;
  time_t delta;
  int isaccurate;		/* whether delta is confirmed accurate */
{
  ARG *step;

  /* ** HOST list ** */
  for (step = HOSTarglist; step; step = step->next)
    if (!strncasecmp(step->name, bp->ut_host, UT_HOSTSIZE)) {
      if   (isaccurate) step->oksecs += (long)delta;
      else step->errsecs += (long)delta;
      break ;
    }
  if (!step && noargs) {		/* new host, add to linked list */
    char *s = (char *)malloc(UT_HOSTSIZE + 1);
    strncpy (s, bp->ut_host, UT_HOSTSIZE);
    addarg(HOST_TYPE, s);
    if   (isaccurate) HOSTarglist->oksecs += (long)delta;
    else HOSTarglist->errsecs += (long)delta;
  }
  /* ** USER list ** */
  for (step = USERarglist; step; step = step->next)
    if (!strncmp(step->name, bp->ut_name, UT_NAMESIZE)) {
      if   (isaccurate) step->oksecs += (long)delta;
      else step->errsecs += (long)delta;
      break ;
    }
  if (!step && noargs) {
    char *s = (char *)malloc(UT_NAMESIZE + 1);
    strncpy (s, bp->ut_name, UT_NAMESIZE);
    addarg(USER_TYPE, s);
    if   (isaccurate) USERarglist->oksecs += (long)delta;
    else USERarglist->errsecs += (long)delta;
  }

}	/* end: add_secs() */

/*
 * print the total time for each user supplied argument
 */

pr_totaltimes()
{
  ARG *step;
    
  printf("\nTOTAL TIMES (d+hh:mm) (Total/Err)\n===========\n");
    
  for (step = USERarglist; step; step = step->next)
  {
    time_t tsecs = (step->errsecs + step->oksecs) ;

    printf("USER %s", step->name);
    if (tsecs < SECSPERDAY)
      printf("  (%5.5s)", asctime(gmtime(&tsecs))+11);
    else
      printf(" (%ld+%5.5s)",
	     tsecs / SECSPERDAY, asctime(gmtime(&tsecs))+11);
    printf(" %ld / %ld\n", tsecs, step->errsecs);
  }	/* for() */

  for (step = HOSTarglist; step; step = step->next)
  {
    time_t tsecs = (step->errsecs + step->oksecs) ;

    printf("HOST %s", step->name);
    if (tsecs < SECSPERDAY)
      printf("  (%5.5s)", asctime(gmtime(&tsecs))+11);
    else
      printf(" (%ld+%5.5s)",
	     tsecs / SECSPERDAY, asctime(gmtime(&tsecs))+11);
    printf(" %ld / %ld\n", tsecs, step->errsecs);
  }	/* for() */

}	/* pr_totaltimes() */

/*
 * Print out the entry. Also calculates the delta time (secs the user
 * was logged in
 */

time_t pr_entry(bp, T, isaccurate, crmesg)
  struct tacutmp *bp;
  TTY *T;
  int isaccurate;	/* this login entry and times are accurate */
  char *crmesg;		/* crash message */
{
  char *ct;
  time_t delta;
  int crash = 0;

  if (T->logout < 0)	/* stored as -ve on reboot */
  {
    crash = 1;
    T->logout = -T->logout;
  }
  delta = (T->logout ? T->logout:time(NULL)) - bp->ut_time;

  if (totalsonly)
  {
    if (delta < 0)
      return (0);
    else
      return (delta);
  }
    
  ct = ctime(&bp->ut_time);
  printf("%-*.*s  %-*.*s %-*.*s %10.10s %5.5s ",
	 UT_NAMESIZE, UT_NAMESIZE, bp->ut_name,
	 UT_LINESIZE, UT_LINESIZE, bp->ut_line,
	 UT_HOSTSIZE, UT_HOSTSIZE, bp->ut_host,
	 ct, ct + 11);

  if (!T->logout)
    printf(" online");
  else if (crash)
    printf("- %s", crmesg);	/* shutdown or crash */
  else
    printf("- %5.5s", ctime(&T->logout)+11);

  if (delta < 0) {	/* should not happen */
    printf(" (negative)");
    delta = 0;
  }
  else if (delta < SECSPERDAY)
    printf("  (%5.5s)", asctime(gmtime(&delta))+11);
  else
    printf(" (%ld+%5.5s)", delta / SECSPERDAY, asctime(gmtime(&delta))+11);

  printf(" %ld%s\n", delta, isaccurate ? " ":" inaccurate");

  return (delta);

}	/* end pr_entry() */


/*
 * hostconv --
 *	convert the hostname to search pattern; if the supplied host name
 *	has a domain attached that is the same as the current domain, rip
 *	off the domain suffix since that's what login(1) does.
 */
void
hostconv(arg)
  char *arg;
{
  static int first = 1;
  static char *hostdot, name[MAXHOSTNAMELEN];
  char *argdot;

  if (!(argdot = strchr(arg, '.')))
    return;
  if (first) {
    first = 0;
    if (gethostname(name, sizeof(name))) {
      perror("last: gethostname");
      exit (1);
    }
    hostdot = strchr(name, '.');
  }
  if (hostdot && !strcasecmp(hostdot, argdot))
    *argdot = '\0';
}

/*
 * Following are 'who' routines for printing out the data in the
 * UTMP file to show users currently logged in.
 */

doutmp()
{
  time_t now, delta;
  struct tacutmp u ;
  FILE   *ut;
  u_long l;
  int    nusers = 0, tusers = 0;	/* wanted users, total users */

  now = time(NULL);

  if (file == (char *)NULL || *file == '\0')
    file = UTMP ;
  if ((ut = fopen(file, "r")) == NULL) {
    fprintf(stderr, "%s: %s: ", prognm, file);
    perror ((char *)NULL);
    exit (1);
  }

#define HEADER	"USER     TTY      FROM              LOGIN@  ONLINE(d+h:m)\n"
  (void)printf(HEADER);

  while (fread(&u, sizeof(u), 1, ut)) {
    if (islogout(&u))
      continue;
    ++tusers;
    if (want(&u, NO))  {
      (void)printf("%-*.*s %-*.*s %-*.*s ",
		   UT_NAMESIZE, UT_NAMESIZE, u.ut_name,
		   UT_LINESIZE, UT_LINESIZE, u.ut_line,
		   UT_HOSTSIZE, UT_HOSTSIZE, *(u.ut_host) ? u.ut_host : "-");
      delta = now - u.ut_time;
      pr_attime(&delta, &(u.ut_time), &now); /* */
      if (delta < SECSPERDAY)
	printf("  (%5.5s)",
	       asctime(gmtime(&delta))+11);
      else
	printf(" (%ld+%5.5s)",
	       delta / SECSPERDAY,
	       asctime(gmtime(&delta))+11);
      printf("\n");
      ++nusers;
      continue;
    }

    (void)fclose(ut);
    (void)printf("\nTotal users: %d (%d)\n", nusers, tusers);

  }	/* end: for */

}
/*
 * pr_attime --
 *	Print the time that the user logged in
 */

pr_attime(delta, started, now)
  time_t *delta, *started, *now;
{
  struct tm *tp;

  tp = localtime(started);

  if (*delta > (SECSPERDAY * DAYSPERWEEK))	/* more than a week */
    printf("%2.2d-%2.2d-%2.2d",
	   tp->tm_mday, tp->tm_mon + 1, tp->tm_year);

  else if (*now / SECSPERDAY != *started / SECSPERDAY) {	/* not today */
    printf("%s-%2.2d:%2.2d",
	   weekdays[tp->tm_wday], tp->tm_hour, tp->tm_min);
  }

  else {	/* today */
    printf("%2.2d:%2.2d", tp->tm_hour, tp->tm_min);
  }

}	/* end pr_attime() */

